#!/usr/local/bin/perl
# Chat Touring
# Copyright (c) 1996 by Paul Burchard
# Distributed under the terms of the GNU Library General Public License
#
# This implementation uses a locked log file as a simple, robust
# "server" for managing state and synchronization.

# CONFIG:
#
# Note: the $logfile must be writable by the server; and thus for
# security reasons, should NOT be directly accessible via HTTP.
# This should be an absolute path.
#
$logfile = "\\ChaTour\\chat.log";
#
# For UNIX, these should agree with the constants used by flock(2).
# For NT, just set $unix to zero and ignore the constants.
#
$unix = 0;
$LOCK_EX = 2;
$LOCK_UN = 8;
#
# Put limits on I/O volume to ward off denial-of-service attacks.
#
$maxCharsIn = 1024;
$maxLinesOut = 40;

# REQUEST:
#
# Parse FORM input (values of 'i', 'n', 'r', 's' are used).
%rawData = (); %formData = ();
foreach(split('&', $ENV{QUERY_STRING})) {
	($name, $value) = split('=', $_);
    $rawData{$name} = $value;                
	$value =~ s!\+! !g; $value =~ s!%(..)!pack('c',hex($1))!eg;
    $formData{$name} = $value;                
}

# INPUT:
#
# Append up to $maxCharsIn approved chars (plus automatically prefixed
# serial number, time, and sanitized nickname) to log file.  Do not append
# anything if sanitized input (or nickname) is empty.  Perform file locking
# to prevent corruption.
#
$input = $rawData{'i'};
if(length($input) > $maxCharsIn) { $input = substr($input, 0, $maxCharsIn); }
$input =~ s!([^-_0-9A-Za-z. \+\%])!'%'+sprintf('%02x',unpack('c',$1))!eg;
if(length($nick) > 30) { $nick = substr($nick, 0, 30); }
$nick = $formData{'n'};
$nick =~ y!0-9A-Za-z. +_-%!_!cs;
if(length($input)>0 && length($nick)>0) {
	if(open(LOGFILE, "+>>$logfile")) {
		if($unix) { flock(LOGFILE, $LOCK_EX); }
		$backup = $maxCharsIn+80;
		if(!(seek(LOGFILE, -$backup, 2) && tell(LOGFILE)>0)) {
			seek(LOGFILE, 0, 0);
		}
		$line = 0;
		while(<LOGFILE>) { if(m!^([0-9]*)/!) { $line = $1; } }
		$line++;
		($sec,$min,$hr,$mdy,$mo,$yr,$wdy,$ydy,$isdst) = localtime(time);
		$date = sprintf('%02d:%02d', $hr, $min);
		seek(LOGFILE, 0, 2);
		print LOGFILE "$line/$date $nick: $input\n";
		if($unix) { flock(LOGFILE, $LOCK_UN); }
		close(LOGFILE);
	}
}

# OUTPUT:
#
# Read up to $maxLinesOut lines of approved chars from log file,
# starting at requested line number (or somewhere near the end of the
# log if the request is null).
#
$showLines = $formData{'s'};
$readLine = $formData{'r'};
$output = '';
if(open(READLOG, "<$logfile")) {
	seek(READLOG, 0, 2);
	$backup = $maxCharsIn+80;
	$line = -1; $lineCount = 0;
	while($line<0 || ($line>1 && $line>$readLine)) {
		if(seek(READLOG, -$backup, 1) && tell(READLOG)>0) { $_ = <READLOG>; }
		else { seek(READLOG, 0, 0); $line = 0; }
		if($_ = <READLOG>) {
			if(m!^([0-9]*)/.*\n!) {
				if($line>0 && $line==$1) { $backup *= 2; }
				$line = $1;
				if($readLine <= 0) {
					$readLine = $line - $showLines;
					if($readLine <= 0) { $readLine = 1; }
				}
				if($line == $readLine) { $output = $_; $lineCount = 1; }
			}
		}
	}
	if($line > 0) {
		readLoop: while(<READLOG>) {
			if(m!^([0-9]*)/.*\n!) {
				$line = $1;
				if($line >= $readLine) { $output .= $_; $lineCount++; }
			}
			if($lineCount >= $maxLinesOut) { last readLoop; }
		}
	}
	close(READLOG);
}
$output =~ y!0-9A-Za-z. %+:/\n_-!!cd;

print "Content-type: text/html\n";
print "Expires: Mon, 01 Jan 1900 00:00:00 GMT\n";
print "\n";
print "$output";

1;

